#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________


import os
import re
from coopr.opt.base import *
from coopr.opt.solver import *
import mockmip
import pyutilib.services
import pyutilib.misc
import pyutilib.common
import pyutilib.plugin.core


class GLPK(SystemCallSolver):
    """The GLPK LP/MIP solver
    """

    def __init__(self, **kwds):
        #
        # Call base constructor
        #
        kwds['type'] = 'glpk'
        SystemCallSolver.__init__(self, **kwds)
        #
        # Valid problem formats, and valid results for each format
        #
        self._valid_problem_formats=[ProblemFormat.mod, ProblemFormat.cpxlp, ProblemFormat.mps]
        self._valid_result_formats={}
        self._valid_result_formats[ProblemFormat.mod] = [ResultsFormat.log]
        self._valid_result_formats[ProblemFormat.cpxlp] = [ResultsFormat.log]
        self._valid_result_formats[ProblemFormat.mps] = [ResultsFormat.log]

    def executable(self):
        executable = pyutilib.services.registered_executable("glpsol")
        if executable is None:
            pyutilib.plugin.core.PluginGlobals.env().log.error("Could not locate the 'glpsol' executable, which is required for solver %s" % self.name)
            self.enable = False
            return None
        return executable.get_path()

    def create_command_line(self,executable,problem_files):
        #
        # Define log file
        #
        if self.log_file is None:
           self.log_file = pyutilib.services.TempfileManager.create_tempfile(suffix = '.glpk.log')

        #
        # Define solution file
        #
        if self.soln_file is None:
            self.soln_file = pyutilib.services.TempfileManager.create_tempfile(suffix = '.glpk.soln')

        #
        # Define results file
        #
        if self._results_format is None or self._results_format == ResultsFormat.log:
           self.results_file = self.log_file
           
        #
        # Define command line
        #
        if self._timelimit is not None and self._timelimit > 0.0:
           timing = " --tmlim "+str(self._timelimit)+" "
        else:
           timing = ""
        if self._problem_format == ProblemFormat.cpxlp:
            problem=" --cpxlp " + problem_files[0]
        elif self._problem_format == ProblemFormat.mps:
            problem=" --freemps " + problem_files[0]
        elif self._problem_format == ProblemFormat.mod:
            problem=" --math " + problem_files[0]
            for filename in problem_files[1:]:
                problem += " --data " + filename
        opt = ""
        for key in self.options:
            if isinstance(self.options[key],basestring) and ' ' in self.options[key]:
                opt += " --"+key+" \""+str(self.options[key])+"\""
            else:
                opt += " --"+key+" "+str(self.options[key])
        proc = self._timer + " " + executable + opt + " " + timing + " --output " + self.soln_file + problem
        return pyutilib.misc.Bunch(cmd=proc, log_file=self.log_file, env=None)

    def process_logfile(self):
        """
        Process logfile
        """
        results = SolverResults()
        #
        # Initial values
        #
        results.solver.ncreated=0
        results.solver.nbounded=0
        soln = results.solution.create()
        #
        # Process logfile
        #
        OUTPUT = open(self.log_file)
        output = "".join(OUTPUT.readlines())
        OUTPUT.close()
        #
        # Parse logfile lines
        #
        for line in output.split("\n"):
          tokens = re.split('[ \t]+',line.strip())
          #print "LINE:", tokens
          if len(tokens) > 4 and tokens[1] == "objval":
             soln.value = tokens[3]
          elif len(tokens) > 3 and tokens[0] == "Objective" and tokens[1] == "value":
             soln.value = tokens[3]
          elif len(tokens) > 4 and tokens[0] == "!" and tokens[2] == "objval":
             soln.value = tokens[4]
          elif len(tokens) > 4 and tokens[0] == "+" and tokens[2] == "objval":
             soln.value = tokens[4]
          elif len(tokens) > 4 and tokens[0] == "*" and tokens[2] == "objval":
             soln.value = tokens[4]
          elif len(tokens) > 4 and tokens[0] == "+" and tokens[2] == "mip" and tokens[4] == "not":
             soln.value = "unknown"
             results.problem.lower_bound = tokens[8]
          elif len(tokens) > 4 and tokens[0] == "+" and tokens[1] == "mip" and tokens[4] == "not":
             soln.value = "unknown"
             results.problem.lower_bound = tokens[7]
          elif len(tokens) > 4 and tokens[0] == "+" and tokens[2] == "mip" and tokens[4] != "not":
             soln.value = tokens[4]
             if tokens[6] != "tree":
                results.problem.lower_bound = tokens[6]
          elif len(tokens) > 4 and tokens[0] == "+" and tokens[1] == "mip" and tokens[4] != "not":
             soln.value = tokens[3]
             results.problem.lower_bound = tokens[5]
          elif len(tokens) == 6 and tokens[0] == "OPTIMAL" and tokens[1] == "SOLUTION" and tokens[5] == "PRESOLVER":
             results.solver.ncreated = 0
             results.solver.nbounded = 0
             soln.status = SolutionStatus.optimal
          elif len(tokens) == 7 and tokens[1] == "OPTIMAL" and tokens[2] == "SOLUTION" and tokens[6] == "PRESOLVER":
             results.solver.ncreated = 0
             results.solver.nbounded = 0
             soln.status = SolutionStatus.optimal
          elif len(tokens) > 10 and tokens[0] == "+" and tokens[8] == "empty":
             results.solver.ncreated = tokens[11][:-1]
             results.solver.nbounded = tokens[11][:-1]
          elif len(tokens) > 9 and tokens[0] == "+" and tokens[7] == "empty":
             results.solver.ncreated = tokens[10][:-1]
             results.solver.nbounded = tokens[10][:-1]
          elif len(tokens) == 2 and tokens[0] == "sys":
             results.solver.systime=tokens[1]
          elif len(tokens) == 2 and tokens[0] == "user":
             results.solver.usrtime=tokens[1]
          elif len(tokens) > 2 and tokens[0] == "OPTIMAL" and tokens[1] == "SOLUTION":
             soln.status = SolutionStatus.optimal
          elif len(tokens) > 2 and tokens[0] == "INTEGER" and tokens[1] == "OPTIMAL":
             soln.status = SolutionStatus.optimal
          elif len(tokens) > 2 and tokens[0] == "TIME" and tokens[2] == "EXCEEDED;":
             soln.status = SolutionStatus.stoppedByLimit
        if results.problem.upper_bound == "inf":
           results.problem.upper_bound = "Infinity"
        if results.problem.lower_bound == "-inf":
           results.problem.lower_bound = "-Infinity"
        try:
            tmp = eval(results.problem.upper_bound.strip())
            results.problem.upper_bound = str(tmp)
        except:
            pass
        try:
            tmp = eval(results.problem.lower_bound.strip())
            results.problem.lower_bound = str(tmp)
        except:
            pass
        try:
            tmp = eval(soln.value.strip())
            soln.value = str(tmp)
        except:
            pass
        if soln.status is SolutionStatus.optimal:
           soln.gap=0.0
        elif soln.status is SolutionStatus.stoppedByLimit:
           soln.gap = "Infinity" # until proven otherwise
           if "lower_bound" in dir(results.problem):
              if results.problem.lower_bound is "-Infinity":
                 soln.gap="Infinity"
              elif not results.problem.lower_bound is None:
                 if "upper_bound" not in dir(results.problem):
                    gap="Infinity"
                 elif results.problem.upper_bound is None:
                    gap="Infinity"
                 else:
                    soln.gap=eval(soln.value) - eval(results.problem.lower_bound)
           elif "upper_bound" in dir(results.problem):
              if results.problem.upper_bound is "Infinity":
                 soln.gap="Infinity"
              elif not results.problem.upper_bound is None:
                 soln.gap=eval(results.problem.upper_bound) - eval(soln.value)
        if results.solver.status is SolverStatus.error:
           results.solution.delete(0)
        return results

    def process_other_data(self,results):
        lp_solution=False
        if not os.path.exists(self.soln_file):
           return
        soln = results.solution(0)
        INPUT = open(self.soln_file,"r")
        state=0
        results.problem.num_objectives=1
        for line in INPUT:
          tokens = re.split('[ \t]+',line.strip())
          ##print "LINE",line,len(tokens)
          ##print "TOKENS",tokens
          if state==0:
             #
             # Processing initial header
             #
             if len(tokens) == 2 and tokens[0] == "Problem:":
                results.problem.name = tokens[1]
             elif len(tokens) == 2 and tokens[0] == "Rows:":
                results.problem.num_constraints = tokens[1]
             elif len(tokens) >= 2 and tokens[0] == "Columns:":
                results.problem.num_variables = tokens[1]
             elif len(tokens) == 2 and tokens[0] == "Non-zeros:":
                results.problem.num_nonzeros = tokens[1]
             elif len(tokens) >= 2 and tokens[0] == "Status:":
                if tokens[1] == "OPTIMAL":
                   soln.status = SolutionStatus.optimal
                elif len(tokens) == 3 and tokens[1] == "INTEGER" and tokens[2] == "NON-OPTIMAL":
                   soln.status = SolutionStatus.bestSoFar
                elif len(tokens) == 3 and tokens[1] == "INTEGER" and tokens[2] == "OPTIMAL":
                   soln.status = SolutionStatus.optimal
                elif len(tokens) == 3 and tokens[1] == "INTEGER" and tokens[2] == "UNDEFINED":
                   soln.status = SolutionStatus.unbounded                   
                else:
                   print "GLPK WARNING: unknown status: "+" ".join(tokens[1:])
             elif len(tokens) >= 2 and tokens[0] == "Objective:":
                if tokens[4] == "(MINimum)":
                   results.problem.sense = ProblemSense.minimize
                else:
                   results.problem.sense = ProblemSense.maximize
                soln.value=tokens[3]
                if soln.status is SolutionStatus.optimal:
                   if tokens[4] == "(MINimum)":
                        results.problem.lower_bound = soln.value
                        if "upper_bound" in dir(results.problem):
                            del results.problem.upper_bound
                   else:
                        results.problem.upper_bound = soln.value
                        if "lower_bound" in dir(results.problem):
                            del results.problem.lower_bound
             elif len(tokens) > 1 and tokens[0] == "No.":
                if tokens[1] == "Row":
                   state=1
                else:
                   state=2
          elif state==1:
             #
             # Process Constraint Info
             #
             if len(tokens) == 0:
                continue
             if len(tokens) > 1 and tokens[0] == "No.":
                state=2
                if line[65:78].strip() == "Marginal":
                   lp_solution=True
             else:
                tmp=line[0:6]
                tmp = tmp.strip()
                if tmp == "------":
                   continue
                if tmp != "":
                   tmpname = tokens[1]
                   if len(tokens) == 2:
                      continue
                tmp=line[65:78]
                tmp=tmp.strip()
                ##print "HERE",tmp
                if tmp=="":
                   if lp_solution:
                      soln.dual.add( tmpname,0.0 )
                else:
                   if tmp.strip() == "< eps":
                      soln.dual.add( tmpname,0.0 )
                   else:
                      soln.dual.add( tmpname,tmp )
          elif state==2:
             #
             # Process Variable Info
             #
             if len(tokens) == 1:
                state=-1
                continue
             if True:
                tmp=line[0:6]
                tmp = tmp.strip()
                if tmp == "------":
                   continue
                if tmp != "":
                   tmpname = tokens[1]
                   if len(tokens) == 2:
                      continue
                tmp=line[23:36]
                tmp=tmp.strip()
                if tmp=="":
                   soln.variable.add( tmpname,0.0 )
                else:
                   soln.variable.add( tmpname,tmp )
          if state==-1:
             break
        INPUT.close()



class MockGLPK(GLPK,mockmip.MockMIP):
    """A Mock GLPK solver used for testing
    """

    def __init__(self, **kwds):
        try:
           GLPK.__init__(self, **kwds)
        except pyutilib.common.ApplicationError: #pragma:nocover
           pass                        #pragma:nocover
        mockmip.MockMIP.__init__(self,"glpk")

    def available(self, exception_flag=True):
        return GLPK.available(self,exception_flag)

    def create_command_line(self,executable,problem_files):
        command = GLPK.create_command_line(self,executable,problem_files)
        mockmip.MockMIP.create_command_line(self,executable,problem_files)
        return command

    def executable(self):
        return mockmip.MockMIP.executable(self)

    def _execute_command(self,cmd):
        return mockmip.MockMIP._execute_command(self,cmd)

    def _convert_problem(self,args,pformat,valid_pformats):
        if pformat in [ProblemFormat.mps,ProblemFormat.cpxlp]:
           return (args,pformat,None)
        else:
           return (args,ProblemFormat.cpxlp,None)


pyutilib.services.register_executable(name="glpsol")
SolverRegistration("glpk", GLPK)
SolverRegistration("_mock_glpk", MockGLPK)
